import{l as b,m as d,s as v,q as B,z as q,G as g,H as h,K as C,W as A,X as D,t as G,x as y,D as H,J as z,a5 as k,M as J,n as K,v as m,w as c,A as M,N,O,Q,an as R}from"./index.881b9b68.js";function S(n){let e,l,t,s,f;const r=n[7].default,a=J(r,n,n[6],null);return{c(){e=K("button"),a&&a.c(),m(e,"class",l="gr-button gr-button-"+n[3]+" gr-button-"+n[2]+" "+n[4]),m(e,"id",n[0]),c(e,"!hidden",!n[1])},m(i,_){y(i,e,_),a&&a.m(e,null),t=!0,s||(f=M(e,"click",n[8]),s=!0)},p(i,[_]){a&&a.p&&(!t||_&64)&&N(a,r,i,i[6],t?Q(r,i[6],_,null):O(i[6]),null),(!t||_&28&&l!==(l="gr-button gr-button-"+i[3]+" gr-button-"+i[2]+" "+i[4]))&&m(e,"class",l),(!t||_&1)&&m(e,"id",i[0]),_&30&&c(e,"!hidden",!i[1])},i(i){t||(g(a,i),t=!0)},o(i){h(a,i),t=!1},d(i){i&&z(e),a&&a.d(i),s=!1,f()}}}function W(n,e,l){let t,{$$slots:s={},$$scope:f}=e,{style:r={}}=e,{elem_id:a=""}=e,{visible:i=!0}=e,{variant:_="secondary"}=e,{size:u="sm"}=e;function w(o){k.call(this,n,o)}return n.$$set=o=>{"style"in o&&l(5,r=o.style),"elem_id"in o&&l(0,a=o.elem_id),"visible"in o&&l(1,i=o.visible),"variant"in o&&l(2,_=o.variant),"size"in o&&l(3,u=o.size),"$$scope"in o&&l(6,f=o.$$scope)},n.$$.update=()=>{n.$$.dirty&32&&l(4,{classes:t}=R(r,["full_width"]),t)},[a,i,_,u,t,r,f,s,w]}class X extends b{constructor(e){super(),d(this,e,W,S,v,{style:5,elem_id:0,visible:1,variant:2,size:3})}}function j(n){let e=n[5](n[3])+"",l;return{c(){l=G(e)},m(t,s){y(t,l,s)},p(t,s){s&40&&e!==(e=t[5](t[3])+"")&&H(l,e)},d(t){t&&z(l)}}}function E(n){let e,l;return e=new X({props:{variant:n[4],elem_id:n[1],style:n[0],visible:n[2],$$slots:{default:[j]},$$scope:{ctx:n}}}),e.$on("click",n[6]),{c(){B(e.$$.fragment)},m(t,s){q(e,t,s),l=!0},p(t,[s]){const f={};s&16&&(f.variant=t[4]),s&2&&(f.elem_id=t[1]),s&1&&(f.style=t[0]),s&4&&(f.visible=t[2]),s&168&&(f.$$scope={dirty:s,ctx:t}),e.$set(f)},i(t){l||(g(e.$$.fragment,t),l=!0)},o(t){h(e.$$.fragment,t),l=!1},d(t){C(e,t)}}}function F(n,e,l){let t;A(n,D,u=>l(5,t=u));let{style:s={}}=e,{elem_id:f=""}=e,{visible:r=!0}=e,{value:a}=e,{variant:i="primary"}=e;function _(u){k.call(this,n,u)}return n.$$set=u=>{"style"in u&&l(0,s=u.style),"elem_id"in u&&l(1,f=u.elem_id),"visible"in u&&l(2,r=u.visible),"value"in u&&l(3,a=u.value),"variant"in u&&l(4,i=u.variant)},[s,f,r,a,i,t,_]}class I extends b{constructor(e){super(),d(this,e,F,E,v,{style:0,elem_id:1,visible:2,value:3,variant:4})}}var P=I;const T=["static"],U=n=>({type:"string",description:"button label",example_data:n.value||"Run"});export{P as Component,U as document,T as modes};
