import{l as b,m as p,s as h,q as c,z as m,G as _,H as f,K as d,Z as v,$ as C,M as k,o as $,x as q,a1 as w,a2 as S,N as j,O as z,Q as G,J as H,a5 as J}from"./index.881b9b68.js";import{C as K}from"./CarouselItem.svelte_svelte_type_style_lang.4dd79b13.js";function M(l){let s,o,t;const i=[l[2]];let r={};for(let e=0;e<i.length;e+=1)r=v(r,i[e]);s=new C({props:r});const u=l[3].default,n=k(u,l,l[5],null);return{c(){c(s.$$.fragment),o=$(),n&&n.c()},m(e,a){m(s,e,a),q(e,o,a),n&&n.m(e,a),t=!0},p(e,a){const g=a&4?w(i,[S(e[2])]):{};s.$set(g),n&&n.p&&(!t||a&32)&&j(n,u,e,e[5],t?G(u,e[5],a,null):z(e[5]),null)},i(e){t||(_(s.$$.fragment,e),_(n,e),t=!0)},o(e){f(s.$$.fragment,e),f(n,e),t=!1},d(e){d(s,e),e&&H(o),n&&n.d(e)}}}function N(l){let s,o;return s=new K({props:{elem_id:l[0],visible:l[1],$$slots:{default:[M]},$$scope:{ctx:l}}}),s.$on("change",l[4]),{c(){c(s.$$.fragment)},m(t,i){m(s,t,i),o=!0},p(t,[i]){const r={};i&1&&(r.elem_id=t[0]),i&2&&(r.visible=t[1]),i&36&&(r.$$scope={dirty:i,ctx:t}),s.$set(r)},i(t){o||(_(s.$$.fragment,t),o=!0)},o(t){f(s.$$.fragment,t),o=!1},d(t){d(s,t)}}}function O(l,s,o){let{$$slots:t={},$$scope:i}=s,{elem_id:r=""}=s,{visible:u=!0}=s,{loading_status:n}=s;function e(a){J.call(this,l,a)}return l.$$set=a=>{"elem_id"in a&&o(0,r=a.elem_id),"visible"in a&&o(1,u=a.visible),"loading_status"in a&&o(2,n=a.loading_status),"$$scope"in a&&o(5,i=a.$$scope)},[r,u,n,t,e,i]}class Q extends b{constructor(s){super(),p(this,s,O,N,h,{elem_id:0,visible:1,loading_status:2})}}var A=Q;const B=["static"];export{A as Component,B as modes};
