import{l as k,m as A,s as q,n as b,t as y,o as h,q as j,v as g,w as p,x as z,y as d,z as B,A as D,D as G,G as C,H as w,J as H,K as J,M as K,N as M,O as N,Q as O}from"./index.881b9b68.js";import{C as Q}from"./Column.696be89b.js";function S(l){let e;const s=l[6].default,t=K(s,l,l[7],null);return{c(){t&&t.c()},m(n,i){t&&t.m(n,i),e=!0},p(n,i){t&&t.p&&(!e||i&128)&&M(t,s,n,n[7],e?O(s,n[7],i,null):N(n[7]),null)},i(n){e||(C(t,n),e=!0)},o(n){w(t,n),e=!1},d(n){t&&t.d(n)}}}function E(l){let e,s,t,n,i,u,m,r,f,_,o;return r=new Q({props:{visible:l[3],$$slots:{default:[S]},$$scope:{ctx:l}}}),{c(){e=b("div"),s=b("div"),t=b("span"),n=y(l[0]),i=h(),u=b("span"),u.textContent="\u25BC",m=h(),j(r.$$.fragment),g(u,"class","transition"),p(u,"rotate-90",!l[3]),g(s,"class","w-full flex justify-between cursor-pointer"),g(e,"id",l[1]),g(e,"class","p-3 border border-gray-200 dark:border-gray-700 rounded-lg flex flex-col gap-3 hover:border-gray-300 dark:hover:border-gray-600 transition"),p(e,"hidden",!l[2])},m(a,c){z(a,e,c),d(e,s),d(s,t),d(t,n),d(s,i),d(s,u),d(e,m),B(r,e,null),f=!0,_||(o=D(s,"click",l[4]),_=!0)},p(a,[c]){(!f||c&1)&&G(n,a[0]),c&8&&p(u,"rotate-90",!a[3]);const v={};c&8&&(v.visible=a[3]),c&128&&(v.$$scope={dirty:c,ctx:a}),r.$set(v),(!f||c&2)&&g(e,"id",a[1]),c&4&&p(e,"hidden",!a[2])},i(a){f||(C(r.$$.fragment,a),f=!0)},o(a){w(r.$$.fragment,a),f=!1},d(a){a&&H(e),J(r),_=!1,o()}}}function F(l,e,s){let t,{$$slots:n={},$$scope:i}=e,{label:u}=e,{elem_id:m}=e,{visible:r=!0}=e,{open:f=!0}=e;const _=()=>{s(3,t=!t)};return l.$$set=o=>{"label"in o&&s(0,u=o.label),"elem_id"in o&&s(1,m=o.elem_id),"visible"in o&&s(2,r=o.visible),"open"in o&&s(5,f=o.open),"$$scope"in o&&s(7,i=o.$$scope)},l.$$.update=()=>{l.$$.dirty&32&&s(3,t=f)},[u,m,r,t,_,f,n,i]}class I extends k{constructor(e){super(),A(this,e,F,E,q,{label:0,elem_id:1,visible:2,open:5})}}var R=I;const T=["static"];export{R as Component,T as modes};
