import{l as r,m as h,s as g,M as c,n as d,v as o,w as f,x as v,N as q,O as w,Q as b,G as C,H as R,J as G}from"./index.881b9b68.js";function H(i){let e,a;const u=i[5].default,t=c(u,i,i[4],null);return{c(){e=d("div"),t&&t.c(),o(e,"class","flex row w-full flex-wrap gap-4"),o(e,"id",i[1]),f(e,"gr-compact",i[3]==="compact"),f(e,"gr-panel",i[3]==="panel"),f(e,"unequal-height",i[0].equal_height===!1),f(e,"items-stretch",i[0].equal_height),f(e,"!hidden",!i[2])},m(l,s){v(l,e,s),t&&t.m(e,null),a=!0},p(l,[s]){t&&t.p&&(!a||s&16)&&q(t,u,l,l[4],a?b(u,l[4],s,null):w(l[4]),null),(!a||s&2)&&o(e,"id",l[1]),s&8&&f(e,"gr-compact",l[3]==="compact"),s&8&&f(e,"gr-panel",l[3]==="panel"),s&1&&f(e,"unequal-height",l[0].equal_height===!1),s&1&&f(e,"items-stretch",l[0].equal_height),s&4&&f(e,"!hidden",!l[2])},i(l){a||(C(t,l),a=!0)},o(l){R(t,l),a=!1},d(l){l&&G(e),t&&t.d(l)}}}function J(i,e,a){let{$$slots:u={},$$scope:t}=e,{style:l={}}=e,{elem_id:s}=e,{visible:_=!0}=e,{variant:m="default"}=e;return i.$$set=n=>{"style"in n&&a(0,l=n.style),"elem_id"in n&&a(1,s=n.elem_id),"visible"in n&&a(2,_=n.visible),"variant"in n&&a(3,m=n.variant),"$$scope"in n&&a(4,t=n.$$scope)},[l,s,_,m,t,u]}class M extends r{constructor(e){super(),h(this,e,J,H,g,{style:0,elem_id:1,visible:2,variant:3})}}var O=M;const Q=["static"];export{O as Component,Q as modes};
