import{l as w,m as q,s as C,V as B,q as d,z as g,G as v,H as h,K as b,Y as S,Z as D,$ as E,o as G,n as j,v as c,w as o,x as k,a1 as J,a2 as K,J as H,a5 as V,aa as M}from"./index.881b9b68.js";function Y(n){let e;return{c(){e=j("div"),c(e,"class","output-html"),c(e,"id",n[0]),o(e,"min-h-[6rem]",n[3]),o(e,"!hidden",!n[2])},m(t,a){k(t,e,a),e.innerHTML=n[1]},p(t,[a]){a&2&&(e.innerHTML=t[1]),a&1&&c(e,"id",t[0]),a&8&&o(e,"min-h-[6rem]",t[3]),a&4&&o(e,"!hidden",!t[2])},i:M,o:M,d(t){t&&H(e)}}}function Z(n,e,t){let{elem_id:a=""}=e,{value:l}=e,{visible:u=!0}=e,{min_height:r=!1}=e;const _=S();return n.$$set=m=>{"elem_id"in m&&t(0,a=m.elem_id),"value"in m&&t(1,l=m.value),"visible"in m&&t(2,u=m.visible),"min_height"in m&&t(3,r=m.min_height)},n.$$.update=()=>{n.$$.dirty&2&&_("change")},[a,l,u,r]}class A extends w{constructor(e){super(),q(this,e,Z,Y,C,{elem_id:0,value:1,visible:2,min_height:3})}}function F(n){var e;let t,a,l,u,r;const _=[n[3],{variant:"center"}];let m={};for(let i=0;i<_.length;i+=1)m=D(m,_[i]);return t=new E({props:m}),u=new A({props:{min_height:n[3]&&((e=n[3])==null?void 0:e.status)!=="complete",value:n[2],elem_id:n[0],visible:n[1]}}),u.$on("change",n[5]),{c(){var i;d(t.$$.fragment),a=G(),l=j("div"),d(u.$$.fragment),c(l,"class","transition"),o(l,"opacity-20",((i=n[3])==null?void 0:i.status)==="pending")},m(i,s){g(t,i,s),k(i,a,s),k(i,l,s),g(u,l,null),r=!0},p(i,s){var T,L;const z=s&8?J(_,[K(i[3]),_[1]]):{};t.$set(z);const f={};s&8&&(f.min_height=i[3]&&((T=i[3])==null?void 0:T.status)!=="complete"),s&4&&(f.value=i[2]),s&1&&(f.elem_id=i[0]),s&2&&(f.visible=i[1]),u.$set(f),s&8&&o(l,"opacity-20",((L=i[3])==null?void 0:L.status)==="pending")},i(i){r||(v(t.$$.fragment,i),v(u.$$.fragment,i),r=!0)},o(i){h(t.$$.fragment,i),h(u.$$.fragment,i),r=!1},d(i){b(t,i),i&&H(a),i&&H(l),b(u)}}}function I(n){let e,t;return e=new B({props:{visible:n[1],elem_id:n[0],disable:!0,$$slots:{default:[F]},$$scope:{ctx:n}}}),{c(){d(e.$$.fragment)},m(a,l){g(e,a,l),t=!0},p(a,[l]){const u={};l&2&&(u.visible=a[1]),l&1&&(u.elem_id=a[0]),l&143&&(u.$$scope={dirty:l,ctx:a}),e.$set(u)},i(a){t||(v(e.$$.fragment,a),t=!0)},o(a){h(e.$$.fragment,a),t=!1},d(a){b(e,a)}}}function N(n,e,t){let{label:a}=e,{elem_id:l=""}=e,{visible:u=!0}=e,{value:r=""}=e,{loading_status:_}=e;const m=S();function i(s){V.call(this,n,s)}return n.$$set=s=>{"label"in s&&t(4,a=s.label),"elem_id"in s&&t(0,l=s.elem_id),"visible"in s&&t(1,u=s.visible),"value"in s&&t(2,r=s.value),"loading_status"in s&&t(3,_=s.loading_status)},n.$$.update=()=>{n.$$.dirty&16&&m("change")},[l,u,r,_,a,i]}class O extends w{constructor(e){super(),q(this,e,N,I,C,{label:4,elem_id:0,visible:1,value:2,loading_status:3})}}var Q=O;const R=["static"],U=n=>({type:"string",description:"HTML output"});export{Q as Component,U as document,R as modes};
