import{l as m,m as b,s as g,a8 as h,ae as p,q as v,z as T,af as k,G as f,H as u,K as q,Y as C,M as w,N as z,O as D,Q as E,a5 as G}from"./index.881b9b68.js";import{T as H}from"./Tabs.6cd99b38.js";function K(s){let e;const l=s[3].default,t=w(l,s,s[6],null);return{c(){t&&t.c()},m(n,_){t&&t.m(n,_),e=!0},p(n,_){t&&t.p&&(!e||_&64)&&z(t,l,n,n[6],e?E(l,n[6],_,null):D(n[6]),null)},i(n){e||(f(t,n),e=!0)},o(n){u(t,n),e=!1},d(n){t&&t.d(n)}}}function M(s){let e,l,t;function n(a){s[4](a)}let _={visible:s[1],elem_id:s[2],$$slots:{default:[K]},$$scope:{ctx:s}};return s[0]!==void 0&&(_.selected=s[0]),e=new H({props:_}),h.push(()=>p(e,"selected",n)),e.$on("change",s[5]),{c(){v(e.$$.fragment)},m(a,c){T(e,a,c),t=!0},p(a,[c]){const o={};c&2&&(o.visible=a[1]),c&4&&(o.elem_id=a[2]),c&64&&(o.$$scope={dirty:c,ctx:a}),!l&&c&1&&(l=!0,o.selected=a[0],k(()=>l=!1)),e.$set(o)},i(a){t||(f(e.$$.fragment,a),t=!0)},o(a){u(e.$$.fragment,a),t=!1},d(a){q(e,a)}}}function N(s,e,l){let{$$slots:t={},$$scope:n}=e;const _=C();let{visible:a=!0}=e,{elem_id:c=""}=e,{selected:o}=e;function r(i){o=i,l(0,o)}function d(i){G.call(this,s,i)}return s.$$set=i=>{"visible"in i&&l(1,a=i.visible),"elem_id"in i&&l(2,c=i.elem_id),"selected"in i&&l(0,o=i.selected),"$$scope"in i&&l(6,n=i.$$scope)},s.$$.update=()=>{s.$$.dirty&1&&_("prop_change",{selected:o})},[o,a,c,t,r,d,n]}class O extends m{constructor(e){super(),b(this,e,N,M,g,{visible:1,elem_id:2,selected:0})}}var Y=O;const j=["static"];export{Y as Component,j as modes};
