import{l as i,m as r,s as o,n as u,t as c,v as d,x as m,y as v,D as f,aa as l,J as p}from"./index.881b9b68.js";function x(n){let e,s;return{c(){e=u("div"),s=c(n[0]),d(e,"class","gr-sample-3d")},m(a,t){m(a,e,t),v(e,s)},p(a,[t]){t&1&&f(s,a[0])},i:l,o:l,d(a){a&&p(e)}}}function _(n,e,s){let{value:a}=e;return n.$$set=t=>{"value"in t&&s(0,a=t.value)},[a]}class D extends i{constructor(e){super(),r(this,e,_,x,o,{value:0})}}var h=D;export{h as E};
