.flowContainer{background-color:"#f6f8fa";overflow:hidden}.flowContainer a{text-decoration:none;color:var(--color-contentPrimary)!important}.flowContainer .port:hover{fill:"#2B65D9"!important;stroke:"#2B65D9"!important}.flowContainer .edge,.flowContainer .marker{stroke:#b1b1b7;stroke-width:1}.flowContainer .edge:hover,.flowContainer .marker:hover{fill:"#2B65D9"!important;stroke:"#2B65D9"!important}.flowContainer .marker{stroke-width:1!important}.flowContainer .edge--dash{stroke:#b1b1b7;stroke-dasharray:5;animation:dashdraw .5s linear infinite;stroke-width:1}@keyframes dashdraw{0%{stroke-dashoffset:10}}.flowContainer foreignObject{overflow:visible}.flowContainer .flow-tools{position:absolute;right:20px;bottom:20px;display:flex;gap:4px}.flowContainer .flow-tools button{color:#8c959f;background:#fff;border:1px solid #8c959f;border-radius:4px;width:24px;height:24px;font-size:16px;line-height:24px;display:table;cursor:pointer}.flowContainer .flow-card{display:flex;align-items:center;justify-content:space-between;gap:8px;height:100%;padding:12px;transition:background-color ease-out .12s,border-color ease-out .12s,box-shadow ease-out .12s,opacity ease-out .12s;border:1px solid rgb(208,215,222);border-radius:6px;cursor:grab;background-color:#fff}.flowContainer .flow-card:hover{box-shadow:0 2px 8px #00000026}.flowContainer .flow-body{display:flex;align-items:center;justify-content:space-between}.flowContainer .flow-content{font-size:14px;font-weight:700}.flowContainer .flow-time{font-size:12px;color:#57606a}.flowContainer .flow-title{position:absolute;left:0;top:-20px;line-height:20px;font-size:12px;color:#57606a}.flowContainer .flow-title:after{position:absolute;top:0;right:0;bottom:0;left:0;content:""}
