class SWEnv:
    instance_uri = "SW_INSTANCE_URI"
    project = "SW_PROJECT"
    task_id = "SW_TASK_ID"
    eval_version = "SW_EVALUATION_VERSION"
    model_version = "SW_MODEL_VERSION"
    runtime_version = "SW_RUNTIME_VERSION"
    status_dir = "SW_TASK_STATUS_DIR"
    log_dir = "SW_TASK_LOG_DIR"
    dataset_uri = "SW_DATASET_URI"
    dataset_row_start = "SW_DATASET_ROW_START"
    dataset_row_end = "SW_DATASET_ROW_END"
    instance_token = "SW_TOKEN"
