import{aQ as s}from"./index.881b9b68.js";const e=["static","dynamic"],o=t=>({type:"string",description:"text string",example_data:t.value||"hello world"});export{s as Component,o as document,e as modes};
