import{l as c,m as u,s as o,n as _,v as l,ab as n,x as f,aa as m,J as g}from"./index.881b9b68.js";function v(i){let e,s;return{c(){e=_("img"),l(e,"class","gr-sample-image object-contain h-20 w-20"),n(e.src,s=i[0])||l(e,"src",s)},m(a,t){f(a,e,t)},p(a,[t]){t&1&&!n(e.src,s=a[0])&&l(e,"src",s)},i:m,o:m,d(a){a&&g(e)}}}function d(i,e,s){let{value:a}=e,{samples_dir:t}=e;return i.$$set=r=>{"value"in r&&s(0,a=r.value),"samples_dir"in r&&s(1,t=r.samples_dir)},[a,t]}class p extends c{constructor(e){super(),u(this,e,d,v,o,{value:0,samples_dir:1})}}var h=p;export{h as E};
