import{l as h,m as g,s as m,a9 as a,v as e,x as d,y as i,aa as l,J as u}from"./index.881b9b68.js";function x(c){let t,r,s,n;return{c(){t=a("svg"),r=a("rect"),s=a("circle"),n=a("polyline"),e(r,"x","3"),e(r,"y","3"),e(r,"width","18"),e(r,"height","18"),e(r,"rx","2"),e(r,"ry","2"),e(s,"cx","8.5"),e(s,"cy","8.5"),e(s,"r","1.5"),e(n,"points","21 15 16 10 5 21"),e(t,"xmlns","http://www.w3.org/2000/svg"),e(t,"width","100%"),e(t,"height","100%"),e(t,"viewBox","0 0 24 24"),e(t,"fill","none"),e(t,"stroke","currentColor"),e(t,"stroke-width","1.5"),e(t,"stroke-linecap","round"),e(t,"stroke-linejoin","round"),e(t,"class","feather feather-image")},m(o,p){d(o,t,p),i(t,r),i(t,s),i(t,n)},p:l,i:l,o:l,d(o){o&&u(t)}}}class w extends h{constructor(t){super(),g(this,t,null,x,m,{})}}export{w as I};
