import{l as r,m as v,s as h,M as w,n as b,v as u,U as g,w as _,x as C,N as q,O as G,Q as H,G as J,H as M,J as N}from"./index.881b9b68.js";function O(i){let e,t,m,f;const o=i[7].default,a=w(o,i,i[6],null);return{c(){e=b("div"),a&&a.c(),u(e,"id",i[2]),u(e,"class",t="overflow-hidden flex flex-col relative col "+g(i[5])),u(e,"style",m=`min-width: min(${i[1]}px, 100%); flex-grow: ${i[0]}`),_(e,"gap-4",i[5].gap!==!1),_(e,"gr-compact",i[4]==="compact"),_(e,"gr-panel",i[4]==="panel"),_(e,"!hidden",!i[3])},m(l,s){C(l,e,s),a&&a.m(e,null),f=!0},p(l,[s]){a&&a.p&&(!f||s&64)&&q(a,o,l,l[6],f?H(o,l[6],s,null):G(l[6]),null),(!f||s&4)&&u(e,"id",l[2]),(!f||s&32&&t!==(t="overflow-hidden flex flex-col relative col "+g(l[5])))&&u(e,"class",t),(!f||s&3&&m!==(m=`min-width: min(${l[1]}px, 100%); flex-grow: ${l[0]}`))&&u(e,"style",m),s&32&&_(e,"gap-4",l[5].gap!==!1),s&48&&_(e,"gr-compact",l[4]==="compact"),s&48&&_(e,"gr-panel",l[4]==="panel"),s&40&&_(e,"!hidden",!l[3])},i(l){f||(J(a,l),f=!0)},o(l){M(a,l),f=!1},d(l){l&&N(e),a&&a.d(l)}}}function Q(i,e,t){let{$$slots:m={},$$scope:f}=e,{scale:o=1}=e,{min_width:a=0}=e,{elem_id:l=""}=e,{visible:s=!0}=e,{variant:c="default"}=e,{style:d={}}=e;return i.$$set=n=>{"scale"in n&&t(0,o=n.scale),"min_width"in n&&t(1,a=n.min_width),"elem_id"in n&&t(2,l=n.elem_id),"visible"in n&&t(3,s=n.visible),"variant"in n&&t(4,c=n.variant),"style"in n&&t(5,d=n.style),"$$scope"in n&&t(6,f=n.$$scope)},[o,a,l,s,c,d,f,m]}class S extends r{constructor(e){super(),v(this,e,Q,O,h,{scale:0,min_width:1,elem_id:2,visible:3,variant:4,style:5})}}var j=S;export{j as C};
