import{l as g,m as h,s as C,q as T,z as k,G as m,H as c,K as p,M as q,N as w,O as z,Q as M,a5 as B,n as D,v as _,am as b,x as E,J as G,Y as H,ar as I,W as J,aC as K,az as N}from"./index.881b9b68.js";import{a as O}from"./Tabs.6cd99b38.js";import{C as Q}from"./Column.696be89b.js";function W(a){let n;const s=a[5].default,t=q(s,a,a[6],null);return{c(){t&&t.c()},m(e,l){t&&t.m(e,l),n=!0},p(e,l){t&&t.p&&(!n||l&64)&&w(t,s,e,e[6],n?M(s,e[6],l,null):z(e[6]),null)},i(e){n||(m(t,e),n=!0)},o(e){c(t,e),n=!1},d(e){t&&t.d(e)}}}function Y(a){let n,s,t;return s=new Q({props:{$$slots:{default:[W]},$$scope:{ctx:a}}}),{c(){n=D("div"),T(s.$$.fragment),_(n,"id",a[0]),_(n,"class","tabitem p-2 border-2 border-t-0 border-gray-200 relative flex"),b(n,"display",a[2]===a[1]?"block":"none",!1)},m(e,l){E(e,n,l),k(s,n,null),t=!0},p(e,[l]){const f={};l&64&&(f.$$scope={dirty:l,ctx:e}),s.$set(f),(!t||l&1)&&_(n,"id",e[0]),l&6&&b(n,"display",e[2]===e[1]?"block":"none",!1)},i(e){t||(m(s.$$.fragment,e),t=!0)},o(e){c(s.$$.fragment,e),t=!1},d(e){e&&G(n),p(s)}}}function j(a,n,s){let t,{$$slots:e={},$$scope:l}=n,{elem_id:f=""}=n,{name:r}=n,{id:u={}}=n;const i=H(),{register_tab:S,unregister_tab:A,selected_tab:d}=I(O);return J(a,d,o=>s(2,t=o)),S({name:r,id:u}),K(()=>()=>A({name:r,id:u})),a.$$set=o=>{"elem_id"in o&&s(0,f=o.elem_id),"name"in o&&s(4,r=o.name),"id"in o&&s(1,u=o.id),"$$scope"in o&&s(6,l=o.$$scope)},a.$$.update=()=>{a.$$.dirty&6&&t===u&&N().then(()=>i("select"))},[f,u,t,d,r,e,l]}class v extends g{constructor(n){super(),h(this,n,j,Y,C,{elem_id:0,name:4,id:1})}}function F(a){let n;const s=a[3].default,t=q(s,a,a[5],null);return{c(){t&&t.c()},m(e,l){t&&t.m(e,l),n=!0},p(e,l){t&&t.p&&(!n||l&32)&&w(t,s,e,e[5],n?M(s,e[5],l,null):z(e[5]),null)},i(e){n||(m(t,e),n=!0)},o(e){c(t,e),n=!1},d(e){t&&t.d(e)}}}function L(a){let n,s;return n=new v({props:{elem_id:a[0],name:a[1],id:a[2],$$slots:{default:[F]},$$scope:{ctx:a}}}),n.$on("select",a[4]),{c(){T(n.$$.fragment)},m(t,e){k(n,t,e),s=!0},p(t,[e]){const l={};e&1&&(l.elem_id=t[0]),e&2&&(l.name=t[1]),e&4&&(l.id=t[2]),e&32&&(l.$$scope={dirty:e,ctx:t}),n.$set(l)},i(t){s||(m(n.$$.fragment,t),s=!0)},o(t){c(n.$$.fragment,t),s=!1},d(t){p(n,t)}}}function P(a,n,s){let{$$slots:t={},$$scope:e}=n,{elem_id:l=""}=n,{label:f}=n,{id:r}=n;function u(i){B.call(this,a,i)}return a.$$set=i=>{"elem_id"in i&&s(0,l=i.elem_id),"label"in i&&s(1,f=i.label),"id"in i&&s(2,r=i.id),"$$scope"in i&&s(5,e=i.$$scope)},[l,f,r,t,u,e]}class R extends g{constructor(n){super(),h(this,n,P,L,C,{elem_id:0,label:1,id:2})}}var Z=R;const y=["static"];export{Z as Component,y as modes};
