import{l as u,m as _,s as c,V as m,q as d,z as b,G as f,H as r,K as g,M as p,N as v,O as $,Q as h}from"./index.881b9b68.js";function k(o){let s;const l=o[2].default,e=p(l,o,o[3],null);return{c(){e&&e.c()},m(t,n){e&&e.m(t,n),s=!0},p(t,n){e&&e.p&&(!s||n&8)&&v(e,l,t,t[3],s?h(l,t[3],n,null):$(t[3]),null)},i(t){s||(f(e,t),s=!0)},o(t){r(e,t),s=!1},d(t){e&&e.d(t)}}}function B(o){let s,l;return s=new m({props:{elem_id:o[0],visible:o[1],explicit_call:!0,$$slots:{default:[k]},$$scope:{ctx:o}}}),{c(){d(s.$$.fragment)},m(e,t){b(s,e,t),l=!0},p(e,[t]){const n={};t&1&&(n.elem_id=e[0]),t&2&&(n.visible=e[1]),t&8&&(n.$$scope={dirty:t,ctx:e}),s.$set(n)},i(e){l||(f(s.$$.fragment,e),l=!0)},o(e){r(s.$$.fragment,e),l=!1},d(e){g(s,e)}}}function q(o,s,l){let{$$slots:e={},$$scope:t}=s,{elem_id:n}=s,{visible:a=!0}=s;return o.$$set=i=>{"elem_id"in i&&l(0,n=i.elem_id),"visible"in i&&l(1,a=i.visible),"$$scope"in i&&l(3,t=i.$$scope)},[n,a,e,t]}class C extends u{constructor(s){super(),_(this,s,q,B,c,{elem_id:0,visible:1})}}var z=C;const G=["static"];export{z as Component,G as modes};
