class FileTypeError(Exception):
    pass


class FileFormatError(Exception):
    pass


class NoSupportError(Exception):
    pass


class NotFoundError(Exception):
    pass


class SWObjNameFormatError(Exception):
    pass


class UnExpectedConfigFieldError(Exception):
    pass


class URIFormatError(Exception):
    pass


class ConfigFormatError(Exception):
    pass


class ExistedError(Exception):
    pass


class MissingFieldError(Exception):
    pass


class FormatError(Exception):
    pass


class EnvironmentError(Exception):
    pass


class PythonEnvironmentError(Exception):
    pass


class FieldTypeOrValueError(Exception):
    pass


class ParameterError(Exception):
    pass


class ExclusiveArgsError(Exception):
    pass


class InvalidObjectName(Exception):
    pass


class MissingDependencyError(Exception):
    pass
