import{l as _,m as r,s as d,M as m,n as c,v as u,w as f,x as v,N as g,O as b,Q as h,G,H as C,J as p}from"./index.881b9b68.js";function q(n){let s,l;const o=n[3].default,t=m(o,n,n[2],null);return{c(){s=c("div"),t&&t.c(),u(s,"id",n[0]),u(s,"class","svelte-10ogue4"),f(s,"hidden",!n[1])},m(e,i){v(e,s,i),t&&t.m(s,null),l=!0},p(e,[i]){t&&t.p&&(!l||i&4)&&g(t,o,e,e[2],l?h(o,e[2],i,null):b(e[2]),null),(!l||i&1)&&u(s,"id",e[0]),i&2&&f(s,"hidden",!e[1])},i(e){l||(G(t,e),l=!0)},o(e){C(t,e),l=!1},d(e){e&&p(s),t&&t.d(e)}}}function w(n,s,l){let{$$slots:o={},$$scope:t}=s,{elem_id:e=""}=s,{visible:i=!0}=s;return n.$$set=a=>{"elem_id"in a&&l(0,e=a.elem_id),"visible"in a&&l(1,i=a.visible),"$$scope"in a&&l(2,t=a.$$scope)},[e,i,t,o]}class H extends _{constructor(s){super(),r(this,s,w,q,d,{elem_id:0,visible:1})}}var M=H;const N=["static"];export{M as Component,N as modes};
